/*
  bootcode.ino - Update Boot-Loader

  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include "bootcode.h"
#include "nvmctrl.h"

void setup()
{
}

void loop()
{
  const uint16_t *data;
  uint16_t addr;
  /* disable interrupt (required) */
  cli();
  /* erase boot code */
  addr = FLASHEND + 1;
  do {
    addr -= SPM_ERASE_PAGESIZE;
    boot_page_erase_safe(addr);
  } while (addr > BOOTCODE_ADDR);
  // erase app entry
  boot_page_erase_safe(addr - SPM_ERASE_PAGESIZE);
  /* write boot code */
  data = (const uint16_t *)BOOTCODE_DATA;
  do {
    boot_page_fill_safe(addr, pgm_read_word(data++));
    addr += sizeof(*data);
    if ((addr & (SPM_PAGESIZE - 1)) == 0)
      boot_page_write_safe(addr - SPM_PAGESIZE);
  } while (addr < sizeof(BOOTCODE_DATA) + BOOTCODE_ADDR);
  /* erase boot vector */
  addr = 0;
  boot_page_erase_safe(addr);
  /* write boot vector */
  data = (const uint16_t *)BOOTCODE_VECT;
  do {
    boot_page_fill_safe(addr, *data++);
    addr += sizeof(*data);
    if ((addr & (SPM_PAGESIZE - 1)) == 0)
      boot_page_write_safe(addr - SPM_PAGESIZE);
  } while (addr < sizeof(BOOTCODE_VECT));
  boot_rww_enable_safe();
  /* run bootloader (watchdog reset) */
  wdt_enable(WDTO_15MS);
  while (1) ;
}
