/*
  nvmctrl.cpp - NVMCTRL Library for AVR Series
  
  Copyright (c) 2022 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#pragma once

#if !defined(SPM_PAGESIZE) && defined(PROGMEM_PAGE_SIZE)
  #define SPM_PAGESIZE PROGMEM_PAGE_SIZE
#endif

#if defined(__AVR_ATtiny841__) || defined(__AVR_ATtiny441__)
  #define SPM_ERASE_PAGESIZE (SPM_PAGESIZE * 4)
  #define ERASE_PAGE_OPT 0x80
#else
  #define SPM_ERASE_PAGESIZE SPM_PAGESIZE
  #define ERASE_PAGE_OPT 0x00
#endif

#if defined(NVMCTRL)
  #define boot_spm_busy()      (NVMCTRL.STATUS & (NVMCTRL_EEBUSY_bm | NVMCTRL_FBUSY_bm))
  #define boot_spm_busy_wait() do {} while(boot_spm_busy())
  #define boot_page_bufclr()                          \
    do {                                              \
	    CCP = CCP_SPM_gc;                               \
	    NVMCTRL.CTRLA = NVMCTRL_CMD_PAGEBUFCLR_gc;      \
    } while (0)
  #define boot_page_fill(address, data)               \
    do {                                              \
	    NVMCTRL.ADDRL = (char)((address) >> 0);         \
	    NVMCTRL.ADDRH = (char)((address) >> 8);         \
  	  NVMCTRL.DATAL = (char)((data) >> 0);            \
	    NVMCTRL.DATAH = (char)((data) >> 8);            \
    } while (0)
  #define boot_page_erase(address)                    \
    do {                                              \
	    boot_page_fill(address, 0xFFFF);                \
      CCP = CCP_SPM_gc;                               \
      NVMCTRL.CTRLA = NVMCTRL_CMD_PAGEERASE_gc;       \
    } while (0)
  #define boot_page_write(address)                    \
    do {                                              \
	    NVMCTRL.ADDRL = (char)((address) >> 0);         \
	    NVMCTRL.ADDRH = (char)((address) >> 8);         \
	    CCP = CCP_SPM_gc;                               \
  	  NVMCTRL.CTRLA = NVMCTRL_CMD_PAGEWRITE_gc;       \
    } while (0)
  #define boot_page_fill_safe(address, data)          \
    do {                                              \
	    boot_spm_busy_wait();                           \
	    boot_page_fill(address, data);                  \
    } while (0)
  #define boot_page_erase_safe(address)               \
    do {                                              \
	    boot_spm_busy_wait();                           \
	    boot_page_erase (address);                      \
    } while (0)
  #define boot_page_write_safe(address)               \
    do {                                              \
	    boot_spm_busy_wait();                           \
	    boot_page_write (address);                      \
    } while (0)
  #define boot_rww_enable_safe()
#else
  #include <avr/boot.h>
  #if defined(CTPB)
    #define boot_page_bufclr()                        \
      do {                                            \
        __SPM_REG = _BV(__SPM_ENABLE) | _BV(CTPB);    \
	      asm volatile ("spm");                         \
      } while (0)
  #elif defined(RWWSRE)
    #define boot_page_bufclr()                        \
      do {                                            \
        __SPM_REG = _BV(__SPM_ENABLE) | _BV(RWWSRE);  \
	      asm volatile ("spm");                         \
      } while (0)
  #else
    #define boot_page_bufclr()                        \
      do {                                            \
        __SPM_REG = _BV(__SPM_ENABLE);                \
        asm volatile ("spm");                         \
      } while (0)
  #endif
  #if !defined(RWWSRE) && !defined(ASRE)
    #undef  boot_rww_enable_safe
    #define boot_rww_enable_safe()
  #endif
#endif

#define boot_page_bufclr_safe()                       \
  do {                                                \
    boot_spm_busy_wait();                             \
    boot_page_bufclr();                               \
  } while (0)
